import datetime

from networkBuilder import buildCae # , buildCaeLasagne
from weightInitialization import zeros
from optimization import optimize

import theano.tensor as T
import theano
import numpy as np

import logging
from os.path import expanduser
import os
import sys

class CAE: 

    def __init__(self, folder, ds, normalizationFactor=255., mom=0.9, rms_decay=0.9, lr=0.00001, batchSize=32, 
            optimization_type='None', epochs=2, rms_const=0.00001, ims=84, imc=4, allow_input_downcast=True, bufferSize=1000, 
            saveFrequency=5): 
        #self.layout = layout
        os.makedirs(folder)
        ## Configuring logging
        logname = folder + '/run.log'
        logging.basicConfig(filename=logname,
                            filemode='a',
                            format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                            datefmt='%H:%M:%S',
                            level=logging.DEBUG)
        console = logging.StreamHandler()
        formatter = logging.Formatter('%(message)s')
        console.setFormatter(formatter)
        logging.getLogger('').addHandler(console)

        self.logger = logging.getLogger()

        self.folder = folder
        self.ds = ds
        self.epochs = epochs

        self.lr = lr
        self.mom = mom
        self.rms_decay = rms_decay
        self.rms_const = rms_const

        self.batchSize = batchSize
        self.bufferSize = bufferSize
        self.optimization_type = optimization_type

        self.ims = ims
        self.imc = imc

        self.normalizationFactor = normalizationFactor

        self.allow_input_downcast = allow_input_downcast

        self.saveFrequency = saveFrequency

        self.printParameters()
        self.initializeNetwork()

    def printParameters(self): 
        #print "", self.agent 
        self.logger.info( "" )
        self.logger.info( "--- Parameters ---")
        self.logger.info( "  -Folder: " + str(self.folder) )
        self.logger.info( "  -Data set: " + str(self.ds) )
        #self.logger.info( "  -Game Path: " + str(self.gamePath) )
        #self.logger.info( "" )
        #self.logger.info( "  -Number of actions: " + str(self.nofActions) )
        #self.logger.info( "  -Number of steps: " + str(self.steps) )
        self.logger.info( "" )
        self.logger.info( "  -Epochs: " + str(self.epochs) )
        self.logger.info( "  -Batch size: " + str(self.batchSize) )
        self.logger.info( "  -Buffer size: " + str(self.bufferSize) )
        self.logger.info( "  -Learning rate: " + str(self.lr) )
        #self.logger.info( "  -Gamma: " + str(self.gamma) )
        #self.logger.info( "  -DoubleDQN: " + str(self.doubleDQN) )
        #self.logger.info( "  -Clip reward: " + str(self.clip_reward) )
        #self.logger.info( "  -Clip delta (loss function): " + str(self.clip_delta) )
        #self.logger.info( "  -Target network update fequency: " + str(self.targetNetworkUpdateFrequency) )
        #self.logger.info( "  -Replay Memory size: " + str(self.repMem) )
        self.logger.info( "" )
        self.logger.info( "  -Momentum type: " + str(self.optimization_type) )
        self.logger.info( "  -Momentum: " + str(self.mom) )
        self.logger.info( "  -RMSprop decay rate: " + str(self.rms_decay) )
        self.logger.info( "  -RMSprop constant (epsilon): " + str(self.rms_const) )
        self.logger.info( "" )
        #self.logger.info( "  -Training start frame: " + str(self.trainStartFrame) )
        #self.logger.info( "  -Initial exploration rate: " + str(self.initialExploration) )
        #self.logger.info( "  -Final exploration rate: " + str(self.finalExploration) )
        #self.logger.info( "  -Final exploration frame: " + str(self.finalExplorationFrame) )
        #self.logger.info( "" )
        #self.logger.info( "  -Batch accumilator (loss function): " + str(self.batch_acc) )
        self.logger.info( "  -Normalization factor (for images): " + str(self.normalizationFactor) )
        self.logger.info( "  -Image size (the input size for the CNN): " + str(self.ims) )
        self.logger.info( "  -Image chanel (the number of frames stacked to make the CNN input): " + str(self.imc) )
        self.logger.info( "" )
        #self.logger.info( "  -weightStatusIntervall: " + str(self.weightStatusIntervall) )
        #self.logger.info( "  -Simple network: " + str(self.simpleNetwork) )
        #self.logger.info( "  -Profiling: " + str(self.profile) )
        self.logger.info( "  -Allow downcast: " + str(self.allow_input_downcast) )
        #self.logger.info( "" )
        self.logger.info( "  -Save network frequency: " + str(self.saveFrequency) )
        #self.logger.info( "  -load network from: " + str(self.loadNetwork) )
        self.logger.info( "" )

    def initializeNetwork(self): 
        ## Setting up shared variable on the gpu
        ims = 84
        imc = 4
        b = self.batchSize

        batchIndex = T.iscalar('batchIndex')
        # self.sh_phi = zeros(name='phi', shape=(self.bufferSize, imc, ims, ims)) 
        self.sh_phi = zeros(name='phi', shape=(self.batchSize, imc, ims, ims)) 
        #ds_phi = T.ftensor4('phi')
        phi = self.sh_phi/self.normalizationFactor

        output, theta, temp = buildCae(phi, self.batchSize)
        self.theta = theta

        # temp = theta[-2].output
        loss = output.loss()
        # loss = -loss
        grads = [T.grad(loss, p) for p in theta]
        # pdates = self.getUpdates(grads)
        updates = optimize(theta, grads, self.optimization_type, lr=self.lr, mom=self.mom, rms_decay=self.rms_decay, rms_const=self.rms_const)
        givens = [(self.sh_phi, self.sh_phi[b*batchIndex:b*(batchIndex+1)])]

        print "updates:", updates

        self.logger.info('Compiling training function')
        trainFn = theano.function(
            inputs=[batchIndex],
            #inputs=[], 
            outputs=[loss],
            updates=updates,
            givens=givens,
            allow_input_downcast=self.allow_input_downcast,
            name='caeTrainFn')

        org, dec = output.org_dec()
        testFn = theano.function(
            inputs=[],
            outputs=[org, dec],
            allow_input_downcast=self.allow_input_downcast,
            name='caeTestFn')

        self.trainFn = trainFn
        self.testFn = testFn

        if False:
            theano.printing.pydotprint(loss, outfile='pics/loss_cae.png', var_with_name_simple=True)

    def train(self): 
        self.logger.info('Loading training data')
        phi = self.loadDs(self.ds)
        np.random.shuffle(phi)

        self.logger.info('Training data shape: %s' % str(phi.shape))

        trainFn = self.trainFn

        samples = phi.shape[0]
        sBatch = self.batchSize
        rBatch = self.bufferSize/sBatch
        sBuff = self.bufferSize
        rBuff = samples/sBuff
        """ 
        batches = phi.shape[0]/self.batchSize
        b = self.batchSize
        """
        self.logger.info(self.weightsStatus())
        
        self.logger.info('Starting training')

        for epoch in range(self.epochs): 
            if epoch % self.saveFrequency == 0 and epoch != 0: 
                self.save(self.folder + '/cae_e' + str(epoch))
            totCost = 0
            for iBuff in range(rBuff):
                self.sh_phi.set_value(phi[sBuff*iBuff:sBuff*(iBuff+1)])
                for iBatch in range(rBatch-1): 
                    (cost,) = trainFn(iBatch)
                    totCost += cost
            self.logger.info( "Epoch=%i, Cost=%.4f" % (epoch, totCost/(rBuff*rBatch)) ) 
            if np.isnan(cost): 
                self.logger.info( 'Cost == nan -> breaking' )
                break 

        self.logger.info(self.weightsStatus())
        # fname = self.folder + '/cae_weights'
        # self.save(fname)

    def test(self): 
        print "Testing"
        phi = self.loadDs(self.ds)
        indices = np.random.choice(10000, 32, replace=False) 
        test_phi = np.take(phi, indices, axis=0)

        self.sh_phi.set_value(test_phi)
        org, dec = self.testFn()

        
        import matplotlib.pyplot as plt
        org = np.mean(org, axis=1)
        dec = np.mean(dec, axis=1)

        print np.mean(org - dec)

        for i in range(2): 
            plt.imshow(org[i,:,:])
            plt.pause(1)
            plt.imshow(dec[i,:,:])
            plt.pause(1)

    def encodeDecode(self, phi): 
        self.sh_phi.set_value(phi)
        _, dec = self.testFn()
        return dec

    def weightsStatus(self):
        report = "\n--- Weight summary ---\n"
        s = "  -%-6s: min=%0.4f, max=%0.4f, mean=%0.4f, var=%0.4f\n"
        for p in self.theta: 
            vals = p.get_value()
            minV = np.min(vals)
            maxV = np.max(vals)
            meanV = np.mean(vals)
            varV = np.var(vals)
            name = p.name 

            report += (s % (name, minV, maxV, meanV, varV))
        return report 


    def save(self, fname): 
        l = self.theta
        keys = [ str(i) for i in range(len(l)) ]
        parameters = [ p.get_value() for p in l ]
        kwds = dict(zip(keys, parameters)) 
        np.savez(fname, **kwds)

    def load(self, fname): 
        self.logger.info('Loading network from file=%s' % fname)
        npzfile = np.load(fname)
        l = self.theta
        for i, p in enumerate(l): 
            p.set_value(npzfile[str(i)])

    def loadDs(self, fnames): 
        files = [ np.load(fname) for fname in fnames ]
        phis = [ f['phi'] for f in files ]
        return np.concatenate(phis)
        
if __name__ == '__main__':
    args = sys.argv
    largs = len(args)

    if largs >= 3: 
        name = args[1]
        ds = args[2:]
    else: 
        print 'Usage: name/label, ds* '
        sys.exit()

    folder = '/home/torstein/dqn_runs/cae/' + name

    cae = CAE(
        folder = folder, 
        ds = ds,
        lr = 0.001,
        batchSize=32,
        epochs=31,
        # optimization_type='SGD', 
        optimization_type='RMSprop', 
        bufferSize=2048, 
    )
    cae.train()
    cae.save(folder + '/cae_weights')

"""
if __name__ == '__main__':

    today = datetime.datetime.today()
    timestamp = today.strftime('%Y-%m-%d_%H-%M-%S')

    # folder = expanduser('~/dqn_runs/cae_%s' % timestamp)
    # folder ='/tmp/dqn_runs/run%s' % timestamp
    # folder = '/home/torstein/dqn_runs/cae/breakout_qbert_space_road_double_1'
    folder = '/home/torstein/dqn_runs/cae/games_medium'

    cae = CAE(
        folder = folder, 
        ds = ['/media/torstein/data/dqn_runs/breakout/01_01_standard/ds_step2750000.npz', 
              '/media/torstein/data/dqn_runs/qbert/01_01_standard/ds_step4000000.npz', 
              '/media/torstein/data/dqn_runs/space_invaders/01_01_standard/ds_step4875000.npz',
              '/media/torstein/data/dqn_runs/road_runner/01_01_standard/ds_step4000000.npz',
              '/media/torstein/data/dqn_runs/double_dunk/01_01_standard/ds_step4250000.npz'],
        #ds = ['/home/torstein/imagenet/proc_val2012.npz'], 
        lr = 0.001,
        batchSize=32,
        epochs=31,
        # optimization_type='SGD', 
        optimization_type='RMSprop', 
        bufferSize=2048, 
    )
    cae.train()
    cae.save(folder + '/cae_weights')
"""
