__author__ = 'torstein'

from networkBuilder import buildNature, buildPredictionNet
from weightInitialization import zeros
from dqn_agent import DqnAgent
from optimization import optimize

import theano
import theano.tensor as T

import numpy as np

import itertools

from multiprocessing import Pool
from time import time 
import datetime

import logging
from os.path import expanduser
import os
import pickle

from replay_memory import ReplayMemory

class DQN:
    def __init__(self, folder, gamePath, steps=20000010, lr=0.00025, gamma=0.99, batchSize=32, allow_input_downcast=True,
                 rms_decay=0.95, mom=0.95, rms_const=0.01, optimization_type='nature', doubleDQN=True,
                 trainStartStep=50000, initialExploration=1, epsilonDist=np.array(((0.4, 0.1), (0.3, 0.01), (0.3, 0.5))), finalExplorationStep=1000000,
                 repMem=250000, targetNetworkUpdateFrequency=30000, profile=False, normalizationFactor=255.0, clip_delta=1, clip_reward=1,
                 batch_acc='sum', imageSize=84, imageChanels=4, simpleNetwork=False, weightStatusIntervall=100000, logLevel=logging.DEBUG, 
                 saveFrequency=125000, loadNetwork=None, loadNetworkCae=None, testing=False, teachLives=False, trainPredNet=False, sgdFrequency=4, frozzenLevel=-1):
        ## 200 000 000 frames = 50 000 000 steps
        if not os.path.exists(folder):
            os.makedirs(folder)
        elif not testing: 
            raise ValueError("The folder does already exit. Files in it could be overwriten. ")

        #os.makedirs(folder)
        ## Configuring logging
        logType = '/test.log' if testing else '/run.log'
        logname = folder + logType
        logging.basicConfig(filename=logname,
                            filemode='a',
                            format='%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s',
                            datefmt='%H:%M:%S',
                            level=logging.DEBUG)
        console = logging.StreamHandler()
        formatter = logging.Formatter('%(message)s')
        console.setFormatter(formatter)
        logging.getLogger('').addHandler(console)

        self.logger = logging.getLogger()

        self.folder = folder
        self.gamePath = gamePath

        self.agent = DqnAgent(gamePath, ims=imageSize, imc=imageChanels)
        self.nofActions = len(self.agent.actionSet)
        self.steps = steps
        self.batchSize = batchSize
        self.lr = lr
        self.gamma = gamma
        self.doubleDQN = doubleDQN
        self.allow_input_downcast = allow_input_downcast

        self.mom = mom
        self.rms_decay = rms_decay
        self.rms_const = rms_const
        self.optimization_type = optimization_type

        self.trainStartStep = trainStartStep
        self.initialExploration = initialExploration
        self.epsilonDist = epsilonDist
        self.finalExplorationStep = finalExplorationStep

        self.repMem = repMem

        self.targetNetworkUpdateFrequency = targetNetworkUpdateFrequency

        self.profile = profile

        self.normalizationFactor = normalizationFactor

        self.clip_reward = clip_reward
        self.clip_delta = clip_delta
        self.batch_acc = batch_acc

        self.ims = imageSize
        self.imc = imageChanels

        self.simpleNetwork = simpleNetwork
        self.weightStatusIntervall = weightStatusIntervall

        self.saveFrequency = saveFrequency
        self.loadNetwork = loadNetwork
        self.loadNetworkCae = loadNetworkCae

        self.teachLives = teachLives
        self.trainPredNet = trainPredNet

        self.sgdFrequency = sgdFrequency

        self.frozzenLevel = frozzenLevel

        self.printParameters()

        self.initializeNetwork()

        if loadNetwork: 
            self.load(loadNetwork)

        if loadNetworkCae: 
            self.loadCae(loadNetworkCae)

    def printParameters(self): 
        #print "", self.agent 
        self.logger.info( "" )
        self.logger.info( "--- Parameters ---")
        self.logger.info( "  -Folder: " + str(self.folder) )
        self.logger.info( "  -Game Path: " + str(self.gamePath) )
        self.logger.info( "" )
        self.logger.info( "  -Number of actions: " + str(self.nofActions) )
        self.logger.info( "  -Number of steps: " + str(self.steps) )
        self.logger.info( "" )
        self.logger.info( "  -Batch size: " + str(self.batchSize) )
        self.logger.info( "  -Learning rate: " + str(self.lr) )
        self.logger.info( "  -Gamma: " + str(self.gamma) )
        self.logger.info( "  -DoubleDQN: " + str(self.doubleDQN) )
        self.logger.info( "  -Clip reward: " + str(self.clip_reward) )
        self.logger.info( "  -Clip delta (loss function): " + str(self.clip_delta) )
        self.logger.info( "  -Include loss fo life in reward signal: " + str(self.teachLives) )
        self.logger.info( "  -Target network update fequency: " + str(self.targetNetworkUpdateFrequency) )
        self.logger.info( "  -Replay Memory size: " + str(self.repMem) )
        self.logger.info( "  -SGD Frequency updates: " + str(self.sgdFrequency) )
        self.logger.info( "  -Frozzen Level: " + str(self.frozzenLevel) )
        self.logger.info( "" )
        self.logger.info( "  -Momentum: " + str(self.mom) )
        self.logger.info( "  -RMSprop decay rate: " + str(self.rms_decay) )
        self.logger.info( "  -RMSprop constant (epsilon): " + str(self.rms_const) )
        self.logger.info( "  -Momentum type: " + str(self.optimization_type) )
        self.logger.info( "" )
        self.logger.info( "  -Training start step: " + str(self.trainStartStep) )
        self.logger.info( "  -Initial exploration rate: " + str(self.initialExploration) )
        self.logger.info( "  -Final exploration rate: " + str(self.epsilonDist) )
        self.logger.info( "  -Final exploration step: " + str(self.finalExplorationStep) )
        self.logger.info( "" )
        self.logger.info( "  -Batch accumilator (loss function): " + str(self.batch_acc) )
        self.logger.info( "  -Normalization factor (for images): " + str(self.normalizationFactor) )
        self.logger.info( "  -Image size (the input size for the CNN): " + str(self.ims) )
        self.logger.info( "  -Image chanel (the number of frames stacked to make the CNN input): " + str(self.imc) )
        self.logger.info( "" )
        self.logger.info( "  -Train prediction network: " + str(self.trainPredNet) )
        self.logger.info( "  -weightStatusIntervall: " + str(self.weightStatusIntervall) )
        self.logger.info( "  -Simple network: " + str(self.simpleNetwork) )
        self.logger.info( "  -Profiling: " + str(self.profile) )
        self.logger.info( "  -Allow downcast: " + str(self.allow_input_downcast) )
        self.logger.info( "" )
        self.logger.info( "  -Save network frequency: " + str(self.saveFrequency) )
        self.logger.info( "  -load network from: " + str(self.loadNetwork) )
        self.logger.info( "  -load network (CAE) from: " + str(self.loadNetworkCae) )
        self.logger.info( "" )

    def initializeNetwork(self): 
        ims = self.ims
        imc = self.imc
        b = self.batchSize

        ## Setting up shared variables on the gpu
        self.sh_phi = zeros(name='phi', shape=(b, imc, ims, ims)) 
        self.sh_a = zeros(name='a', shape=(b,)) 
        self.sh_r = zeros(name='r', shape=(b,)) 
        self.sh_phi_n = zeros(name='phi_n', shape=(b, imc, ims, ims)) 
        self.sh_t = zeros(name='t', shape=(b,)) 

        ## Converting and preprocessing the variables 
        nf = self.normalizationFactor

        phi = self.sh_phi/nf
        a = T.cast(self.sh_a, 'int8')
        r = self.sh_r
        phi_n = self.sh_phi_n/nf
        t = T.cast(self.sh_t, 'int8')

        # Building the network
        self.output, self.oTheta, self.tTheta = buildNature(phi, a, r, phi_n, t, 
                                                self.gamma, self.nofActions, self.batch_acc, self.batchSize, 
                                                self.clip_reward, self.clip_delta, self.doubleDQN, self.ims, self.imc) 

        output = self.output
        oTheta = self.oTheta
        tTheta = self.tTheta 

        loss = output.loss()
        if self.frozzenLevel > 0: 
            _oTheta = oTheta[self.frozzenLevel:]
        else: 
            _oTheta = oTheta
        grads = [T.grad(loss, p) for p in _oTheta]
        updates = optimize(_oTheta, grads, self.optimization_type, lr=self.lr, mom=self.mom, rms_decay=self.rms_decay, rms_const=self.rms_const)
        #else: 
        #    grads = [T.grad(loss, p) for p in oTheta]
        #    updates = optimize(oTheta, grads, self.optimization_type, lr=self.lr, mom=self.mom, rms_decay=self.rms_decay, rms_const=self.rms_const)
        self.logger.info(str(grads))
        
        profile = theano.compile.ProfileStats() if self.profile else None

        # Initializing the replay memory
        rm = ReplayMemory(self.batchSize, self.repMem, ims=self.ims, imc=self.imc)

        # Compiling functions
        self.logger.info( "Compiling functions" ) 
        trainFn = theano.function(
                    inputs=[],
                    outputs=[loss],
                    updates=updates,
                    allow_input_downcast=self.allow_input_downcast,
                    name='trainFn',
                    profile=profile)

        aFn = theano.function( 
                    inputs=[],
                    outputs=[output.action(), output.actionValueMean()],
                    allow_input_downcast=self.allow_input_downcast,
                    name='actionFn',
                    profile=profile)

        if False:
            theano.printing.pydotprint(loss, outfile='pics/loss.png', var_with_name_simple=True)
            theano.printing.pydotprint(aFn, outfile='pics/aFn.png', var_with_name_simple=True)
            theano.printing.pydotprint(trainFn, outfile='pics/trainFn.png', var_with_name_simple=True)

        self.trainFn = trainFn
        self.aFn = aFn
        self.rm = rm

        if self.trainPredNet: 
            self.logger.info('Building the next state prediction network') 
            pParams = [ (oTheta[i], oTheta[i+1]) for i in range(0, len(oTheta), 2) ] ## gives [ (W,b), (W,b), (W,b), (W,b) ]
            predOutput, predTheta = buildPredictionNet(phi, phi_n, self.batchSize, pParams)

            predLoss = predOutput.loss()
            predGrads = [T.grad(predLoss, p) for p in predTheta]
            predUpdates = optimize(predTheta, predGrads, 'RMSprop', lr=0.000005, mom=0.9, rms_decay=0.9, rms_const=0.00001)

            predTrainFn = theano.function(
                        inputs=[],
                        outputs=[predLoss],
                        updates=predUpdates,
                        allow_input_downcast=self.allow_input_downcast,
                        name='predTrainFn',
                        profile=profile)

            self.predTrainFn = predTrainFn


    def sampleEpsilon(self): 
        if True: 
            prob     = self.epsilonDist[:,0]
            epsilons = self.epsilonDist[:,1]

            index    = np.random.choice(len(prob), 1, p=prob)
            epsilon  = epsilons[index]
            return epsilon
        else: 
            return np.random.uniform(0.05, 0.5)

    def probExplor(self, step, epsilon): 
        if step < self.trainStartStep:
            return 1
        if step > self.finalExplorationStep:
            return epsilon
        else:
            t = max(0, step-self.trainStartStep)
            fe = epsilon
            ie = self.initialExploration
            fef = self.finalExplorationStep
            ief = self.trainStartStep
            p = (fe-ie)/(fef-ief)*(t-2*ie) + ie
            return p

    def shouldUpdateTargetNetwork(self, frame):
        return (frame % self.targetNetworkUpdateFrequency == 0)

    def updateTargetNetwork(self):
        for p, tp in zip(self.oTheta, self.tTheta):
            tp.set_value(p.get_value())

    def shouldTrain(self, step):
        return step > self.batchSize and step > self.trainStartStep and step % self.sgdFrequency == 0

    """
    def __executeAction(self, a): 
        self.agent.act(a)
        r = self.agent.reward()
        phi_n = self.agent.observe()
        t = self.agent.terminal()
        return (r, phi_n, t)


    def __trainBatch(self, step, trainFn, sh_phi, sh_a, sh_r, sh_phi_n, sh_t, rm): 
        if self.shouldTrain(step):
            b_phi, b_a, b_r, b_phi_n, b_t = rm.sample(step)

            sh_phi.set_value(b_phi)
            sh_a.set_value(b_a)
            sh_r.set_value(b_r)
            sh_phi_n.set_value(b_phi_n)
            sh_t.set_value(b_t)

            c = trainFn()
            return c
            # totCost += c[0]
    """

    def train(self):

        trainFn = self.trainFn
        aFn     = self.aFn
        rm      = self.rm
        agent   = self.agent

        self.logger.info( "Starting training" ) 
        startTime = time()

        phi = self.agent.observe()
        totCost = 0
        episode = 0 
        epsilon = self.sampleEpsilon()

        for step in range(self.steps):
            if agent.terminal(): 
                self.logger.info( "Game=%i, Score=%.4f, avg cost=%.4f, step=%i, sampled epsilon=%.3f, time=%f" % (episode, agent.episodeReward(), (4*totCost/agent.getEpisodeFrameNumber()), step, epsilon, (time() - startTime)) ) 
                
                epsilon = self.sampleEpsilon()
                agent.reset()

                totCost = 0
                episode += 1
                
            if self.shouldPrintWeightStatus(step): 
                self.logger.info( self.weightsStatus() )

            if self.shouldSave(step): 
                fname = self.folder + ('/step%i' % step) 
                self.save(fname)

            if self.shouldUpdateTargetNetwork(step):
                self.logger.info( "Updating target network parameters" ) 
                self.updateTargetNetwork()



            ## Choose action
            if np.random.rand() < self.probExplor(step, epsilon): 
                a = np.random.randint(0,self.nofActions)
            else:
                self.sh_phi.set_value(phi)
                a, _ = aFn()

            ## Interact with environment
            agent.act(a)

            r      = agent.reward()
            phi_n  = agent.observe()
            t      = agent.terminal()

            if self.teachLives: 
                r += agent.livesDelta()

            ## Training the model
            if self.shouldTrain(step):
                b_phi, b_a, b_r, b_phi_n, b_t = rm.sample()

                self.sh_phi.set_value(b_phi)
                self.sh_a.set_value(b_a)
                self.sh_r.set_value(b_r)
                self.sh_phi_n.set_value(b_phi_n)
                self.sh_t.set_value(b_t)

                c = trainFn()
                if self.trainPredNet: 
                    self.predTrainFn()
                totCost += c[0]

            ## Adding sample interaction to replay memory
            rm.store(phi, a, r, t)
            phi = phi_n



        endTime = time()
        self.logger.info( "Finnished. Total training time=%f" % (endTime-startTime) ) 
        if self.profile:
            profile.print_summary()

    def shouldPrintWeightStatus(self, step): 
        return (step%self.weightStatusIntervall == 0) and step != 0

    def weightsStatus(self):
        report = "\n--- Weight summary ---\n"
        s = "  -%-6s: min=%0.4f, max=%0.4f, mean=%0.4f, var=%0.4f\n"
        for p in self.oTheta: 
            vals = p.get_value()
            minV = np.min(vals)
            maxV = np.max(vals)
            meanV = np.mean(vals)
            varV = np.var(vals)
            name = p.name 

            report += (s % (name, minV, maxV, meanV, varV))
        return report 


    def shouldSave(self, step): 
        return step % self.saveFrequency == 0


    def save(self, fname): 
        self.logger.info('Saving network to file=%s' % fname)
        l = self.oTheta + self.tTheta
        keys = [ str(i) for i in range(len(l)) ]
        parameters = [ p.get_value() for p in l ]
        kwds = dict(zip(keys, parameters)) 
        np.savez(fname, **kwds )

    def load(self, fname): 
        self.logger.info('Loading network from file=%s' % fname)
        npzfile = np.load(fname)
        l = self.oTheta +  self.tTheta
        for i, p in enumerate(l): 
            p.set_value(npzfile[str(i)])

    def loadCae(self, fname, case=0): 
        self.logger.info('Loading cae network from file=%s' % fname)
        npzfile = np.load(fname)

        l = list(enumerate(self.oTheta))

        if case==0: 
            self.logger.info('Loading the hidden layer weights. ')
            l = l[:-2] ## load hidden layer
        elif case==1: 
            self.logger.info('Not loading the hidden layer weights. ')
            l = l[:-4] ## not load hidden layer
        elif case==2: 
            self.logger.info('Loading filters only')
            l = l[0:-4:2] ## not load hidden layer
        elif case==3: 
            self.logger.info('Loading first layer only')
            l = l[0:2] ## not load hidden layer
        else: 
            raise ValueError('Unrecognized pcae loading case. case=%i' % case)

        for i, p in l: 
            weights = npzfile[str(i)]
            assert(p.get_value(borrow=True).shape == weights.shape, 'Loaded shapes does not match')
            self.logger.info('Loading weight tensor number %i, with shape=%s' % (i, str(p.get_value(borrow=True).shape)))
            p.set_value(weights)
        self.updateTargetNetwork()


    def dsGen(self, testSteps, fname, e=0.05): 
        self.logger.info('Generating dataset. testSteps=%i, fname=%r' % (testSteps, fname))
        assert testSteps <= self.rm.maxSize, "testSteps must be less than the replay memory size"

        aFn   = self.aFn
        rm    = self.rm
        agent = self.agent

        episode = 0 
        phi = agent.observe()
        for step in range(testSteps): 
            if agent.terminal(): 
                self.logger.info( "Test game=%i, Score=%.4f, step=%i" % (episode, agent.episodeReward(), step) )
                
                agent.reset()

                self.phi = agent.observe()
                episode += 1

            ## Choose action
            if np.random.rand() < e: 
                a = np.random.randint(0, self.nofActions)
            else:
                self.sh_phi.set_value(phi)
                a, _ = aFn()

            agent.act(a)
            r      = agent.reward()
            t      = agent.terminal()

            rm.store(phi, a, r, t)

            phi  = agent.observe()

        self.logger.info( "Test game=%i, Score=%.4f, step=%i" % (episode, agent.episodeReward(), step) )

        # fname = self.folder + ('/ds')
        self.logger.info('Storing the replay memory')
        rm.save(fname)

    def test(self, episodes=30, maxSteps=4500, e=0.05, verbose=True): 
        self.logger.info(' -Testing performance with nature test procedure. episodes=%i, maxSteps=%i, epsilon=%f' % (episodes, maxSteps, e))

        aFn   = self.aFn
        agent = self.agent

        episode = 0 
        phi = agent.observe()

        scores = np.zeros((episodes))

        for episode in range(episodes): 
            for step in range(maxSteps): 
                if agent.terminal(): 
                    break

                ## Choose action
                if np.random.rand() < e: 
                    a = np.random.randint(0, self.nofActions)
                else:
                    self.sh_phi.set_value(phi)
                    a, _ = aFn()

                agent.act(a)
                phi  = agent.observe()

            scores[episode] = agent.episodeReward()
            if verbose: 
                self.logger.info( "   +Episode=%i, Score=%.4f, step=%i" % (episode, agent.episodeReward(), step) )

            agent.reset()
            phi = agent.observe()

        avgScore = np.mean(scores)
        self.logger.info(" -Average score=%f\n" % avgScore)
        return avgScore, scores

    def testGui(self, episodes=1, maxSteps=4500, e=0.05, verbose=True, record=None): 
        self.logger.info(' -Testing with gui. episodes=%i, maxSteps=%i, epsilon=%f, record=%s' % (episodes, maxSteps, e, str(record)))
        agent = DqnAgent(self.gamePath, ims=self.ims, imc=self.imc, gui=True, record=record)

        aFn   = self.aFn
        phi = agent.observe()

        for episode in range(episodes): 
            for step in range(maxSteps): 
                if agent.terminal(): 
                    if verbose: 
                        self.logger.info( "   +Episode=%i, Score=%.4f, step=%i" % (episode, agent.episodeReward(), step) )
                    break

                ## Choose action
                if np.random.rand() < e: 
                    a = np.random.randint(0, self.nofActions)
                else:
                    self.sh_phi.set_value(phi)
                    a, _ = aFn()

                agent.act(a)
                phi  = agent.observe()
                
            agent.reset()
            phi = agent.observe()

    def predActions(self, phi): 
        self.sh_phi.set_value(phi)
        a, _ = self.aFn()
        return a

    def loggMessage(self, message): 
        self.logger.info(message)


if __name__ == '__main__':
    from collections import Counter

    ds = '/media/torstein/data/dqn_runs/breakout/01_01_standard/ds_step2750000.npz'
    net = '/home/torstein/dqn_runs/long/breakout/91_01_sgdFreq_long/step9000000.npz'
    gamePath = '/home/torstein/dqn/games/breakout.bin'
    dqn = DQN(
        folder = '/tmp/testing', 
        gamePath = gamePath, 
        steps = 5000010, 
        repMem = 100,
    )
    dqn.load(net)

    phi = np.load(ds) 
    actions = []
    for i in range(0,len(phi),1000): 
        actions.extend(dqn.predActions(phi[0:1000,:,:,:]))

    print Counter(actions).items()

"""
if __name__ == '__main__':
    today = datetime.datetime.today()
    timestamp = today.strftime('%Y-%m-%d_%H-%M-%S')

    game = 'breakout'
    #game = 'montezuma_revenge'
    #game = 'qbert'
    #game = 'road_runner'
    #game = 'double_dunk'
    #game = 'space_invaders'

    #caeFname = None 
    #caeFname = '/media/torstein/data/dqn_runs/cae/breakout_qbert_space_road_double_1/cae_weights.npz'
    caeFname = '/home/torstein/dqn_runs/cae/imagenet_small/cae_e5.npz'
    caeCase = 0

    case = 'test'
    name = '30_pcae_imagenet_small_e5_case0'

    gamePath = expanduser('~/dqn/games/' + game + '.bin')
    # folder = expanduser('~/dqn_runs/%s_%s' % (game, timestamp))
    # folder = '/tmp/dqn_runs/%s_%s' % (game, timestamp)

    if case == 'normal':
        folder = '/home/torstein/dqn_runs/' + game + '/' + name
        dqn = DQN(
            folder = folder, 
            gamePath = gamePath, 
            steps = 5000010, 
            # steps = 10000010, 
            # steps = 2000, 
            # teachLives = True, 
            # epsilonDist = np.array(((0.4, 0.1), (0.3, 0.01), (0.3, 0.5))), 
            # trainStartStep = 500, 
            # repMem = 3000, 
            # repMem = 500000,
            # trainPredNet = True,
            # batch_acc = 'mean',
            # optimization_type = 'RMSprop_momentum'
            # lr=0.0005,
        )
        #dqn.loggMessage('Testing uniform sampling of epsilon. ')
        if caeFname: 
            dqn.loadCae(caeFname, case=caeCase)
        dqn.train()

    if case == 'testSingle': 
        dqn = DQN(
            folder = '/tmp/dqn_runs/TEST_%s_%s' % (game, timestamp),
            gamePath = gamePath, 
            repMem = 100, 
        )
        dqn.load('/home/torstein/dqn_runs/breakout/01_standard/step4000000.npz')
        dqn.test(verbose=True)

    if case == 'test': 
        folder = '/home/torstein/dqn_runs/breakout/11_noRepMem'
        dqn = DQN(
            folder = folder, #'/tmp/dqn_runs/TEST_%s_%s' % (game, timestamp),
            gamePath = gamePath, 
            repMem = 100, 
            testing = True, 
        )
        files = [f for f in os.listdir(folder) if os.path.isfile(os.path.join(folder, f))]
        files = [f for f in files if f.startswith('step') and f.endswith('.npz')]

        def comp(s1,s2): 
            return np.clip(int(s1[4:-4]) - int(s2[4:-4]), -1, 1) 
        files.sort(cmp=comp)

        res = []
        for f in files: 
            dqn.load(os.path.join(folder, f))
            avgScore, scores = dqn.test(verbose=True)
            res.append((f, avgScore, scores))
        dqn.logger.info('All results: ' + str(res))

        resName = folder + '/res.dat'
        with open(resName, 'wb') as f: 
            pickle.dump(res, f)

    if case == 'testGui': 
        dqn = DQN(
            folder = '/tmp/dqn_runs/testGui_%s_%s' % (game, timestamp),
            gamePath = gamePath, 
            repMem = 100, 
        )
        # dqn.load('/home/torstein/dqn_runs/run_2016-02-09_15:04:13/step7000000.npz') ## breakout, score~30
        # dqn.load('/home/torstein/dqn_runs/run_2016-02-11_01:33:49/step20000000.npz') ## road_runner, score~14000
        # dqn.load('/home/torstein/dqn_runs/run_2016-02-14_23:02:07/step20000000.npz') ## q_bert, score~4000
        # dqn.load('/media/torstein/data/dqn_runs/space_invaders_2016-03-01_21-43-57/step125000.npz') ## space invaders, score~2000
        dqn.load('/media/torstein/data/dqn_runs/breakout_epsilonDist_2016-03-08_13-13-01/step1375000.npz') ## breakout with epsilon dist, score~60
        dqn.testGui(1, record='/home/torstein/records/breakout_epsilonDist')

    if case == 'continue': 
        dqn = DQN(
            folder=folder,
            gamePath = gamePath, 
            trainStartStep=500,
            initialExploration=0.1, 
            loadNetwork='/home/torstein/dqn_runs/run_2016-02-09_15:04:13/step7000000.npz'
        )
        dqn.train()

    if case == 'ds_gen': 
        path = '/media/torstein/data/dqn_runs/double_dunk/01_standard/'
        netName = 'step4250000'
        dsSize = 50000
        dqn = DQN(
            folder = '/tmp/dqn_runs/DSgen_' + timestamp, 
            gamePath = gamePath, 
            repMem = dsSize,
            loadNetwork= path + netName + '.npz'
        )
        dqn.dsGen(dsSize, path + 'ds_' + netName)

    if case == 'pretrainedCae': 
        dqn = DQN(
            folder = expanduser('~/dqn_runs/pCae_%s_%s' % (game, timestamp)),
            gamePath = gamePath, 
            # trainStartStep = 25000,
            # finalExplorationStep = 100000,
            steps=5000010,
            # steps=2000,
            loadNetworkCae='/home/torstein/dqn_runs/cae_2016-02-29_14-00-31/cae_weights.npz', ## Train on imagenet val 2012
        )
        dqn.train()
"""