__author__ = 'torstein'

import sys
from random import randrange
from ale_python_interface import ALEInterface
import numpy as np
from scipy.misc import imresize



class DqnAgent():

    def __init__(self, gameFile, seed=-1, gui=False, ims=84, imc=4, record=None):
        if seed == -1: 
            seed = np.random.randint(2**16)

        self.ale = ALEInterface()
        self.ale.setInt('random_seed', seed)
        if gui: self.configureGui()
        if record: self.configureRecord(record)
        self.ale.loadROM(gameFile)

        self.actionSet = self.ale.getMinimalActionSet() 
        self.actionReward = 0
        self.gameRewardValue = 0
        self._livesDelta = 0

        self.actionRepeat=imc
        self.screenDim = (ims,ims)

        dim = self.ale.getScreenDims()
        self.preFrame = np.zeros((dim[1],dim[0]), dtype='uint8')

        # Other convertion weighting: [0.21, 0.72, 0.07]
        # self.grayConvertion = [0.2989, 0.587, 0.114, 0] # 4D because image from ALE is coded with 32 bit -> 8 bits not used
        self.grayConvertion = [0.2989, 0.587, 0.114]

        # action = 0 -> no op
        self.act(0)

    def configureGui(self):
        if sys.platform == 'darwin':
            import pygame
            pygame.init()
            self.ale.setBool('sound', False) # Sound doesn't work on OSX
        elif sys.platform.startswith('linux'):
            self.ale.setBool('sound', True)
        self.ale.setBool('display_screen', True)

    def configureRecord(self, folder): 
        self.ale.setString('record_screen_dir', folder)
        self.ale.setString('record_sound_filename', folder + '/sound.wav')

    def processFrame(self, frame):
        # print "proccessFrame: frame shape:", frame.shape
        """
        ## code works if frame is array of ints
        conv = self.grayConvertion
        view = np.sum( frame.view(np.uint8).reshape(frame.shape + (4,)) * conv, 1)
        resized = imresize(view.reshape((210,160)), self.screenDim)
        return resized[np.newaxis, ...] # Adding a new axis, Shape (84,84) -> (1,84,84)
        """
        gray = np.sum( frame * self.grayConvertion, 2 )

        maxFrame = np.maximum(self.preFrame, gray)
        self.preFrame = gray

        resized = imresize(maxFrame, self.screenDim)
        return resized[np.newaxis, np.newaxis, ...] # Adding a new axises, Shape (84,84) -> (1,1,84,84)

    def act(self, action):
        action = self.actionSet[action] ## Mapping from range(nofActiosn) -> ale.actionset()
        reward = 0
        preLives = self.ale.lives()
        frames = []
        for i in range(self.actionRepeat):
            reward += self.ale.act(action)
            frame = self.ale.getScreenRGB()
            frames.append(self.processFrame(frame))
        self.observations = np.concatenate(frames, axis=1) 
        self.actionReward = reward
        self.gameRewardValue += reward
        self._livesDelta = self.ale.lives() - preLives

    def observe(self):
        return self.observations

    def reward(self):
        return self.actionReward

    def livesDelta(self): 
        return self._livesDelta

    def episodeReward(self):
        return self.gameRewardValue
    """
    def gameOver(self):
        return self.ale.game_over()
    """
    def terminal(self): 
        return 1 if self.ale.game_over() else 0 

    def reset(self):
        self.ale.reset_game()
        self.gameRewardValue = 0
        self._livesDelta = 0
        self.act(0)

    def getMinimalActionSet(self):
        return self.ale.getMinimalActionSet()
    
    """
    def getScreenDim(self):
        return self.screenDim
    """

    def getEpisodeFrameNumber(self): 
        return self.ale.getEpisodeFrameNumber()

    def getFrameNumber(self): 
        return self.ale.getFrameNumber()


if __name__ == '__main__':
    agent = DqnAgent('../games/ASSAULT.BIN')
