import itertools

#from convolutionalLayer import ConvolutionalLayer
#from hiddenLayer import HiddenLayer
#from dqnOutputLayer import DqnOutputLayer

from cae_layers import CaeOutputLayer, CaeConvolutionEncoder, CaeHiddenEncoder, CaeConvolutionDecoder, InversLayer
from dqn_layers import DqnInputLayer, DqnConvolutionalLayer, DqnFlattenLayer, DqnDenseLayer, DqnOutputLayer
import theano.tensor as T

"""
import lasagne 
from lasagne.layers import InputLayer, Conv2DLayer, DenseLayer, InverseLayer 
from lasagne.nonlinearities import rectify
"""

conv1_f = (32, 4, 8, 8)
conv2_f = (64, 32, 4, 4)
conv3_f = (64, 64, 3, 3)

conv1_s = (4,4)
conv2_s = (2,2)
conv3_s = (1,1)

dense_shape = (3136,512)

def buildNature(phi, a, r, phi_n, t, gamma, nofActions, batch_acc, 
            batchSize, clip_reward, clip_delta, doubleDQN, ims, imc):
    
    dqn_in  = DqnInputLayer(phi, phi_n)
    conv1   = DqnConvolutionalLayer(dqn_in, conv1_f, conv1_s, name='conv1')
    conv2   = DqnConvolutionalLayer(conv1, conv2_f, conv2_s, name='conv2')
    conv3   = DqnConvolutionalLayer(conv2, conv3_f, conv3_s, name='conv3')
    temp    = DqnFlattenLayer(conv3)
    dense   = DqnDenseLayer(temp, dense_shape[0], dense_shape[1], name='dense')
    dqn_out = DqnOutputLayer(dense, dense_shape[1], nofActions, batchSize, gamma, 
                             name="out", a=a, r=r, t=t, batch_acc=batch_acc, 
                             clip_delta=clip_delta, clip_reward=clip_reward, 
                             doubleDQN=doubleDQN) 

    layers = [conv1, conv2, conv3, dense, dqn_out]
    oTheta = list( itertools.chain( *[layer.oTheta for layer in layers] ) )
    tTheta = list( itertools.chain( *[layer.tTheta for layer in layers] ) )

    return dqn_out, oTheta, tTheta


def buildCae(phi, batchSize): 
    layers = []
    chanels = 4
    b = batchSize

    ## Encoder
    enc_l1 = CaeConvolutionEncoder(phi, conv1_f, conv1_s, p_corropt=0.3, name='cae_e1')
    enc_l2 = CaeConvolutionEncoder(enc_l1.output, conv2_f, conv2_s, name='cae_e2')
    enc_l3 = CaeConvolutionEncoder(enc_l2.output, conv3_f, conv3_s, name='cae_e3')
    enc_l4 = CaeHiddenEncoder(T.flatten(enc_l3.output, outdim=2), 3136, 512, name='cae_e_hid') 
    enc_extra = CaeHiddenEncoder(enc_l4.output, 512, 18, name='cae_e_extra') 
    enc_out = enc_extra.output

    ## Decoder 
    dec_extra = CaeHiddenEncoder(enc_out, 18, 512, name='cae_d_hid')
    dec_l5 = CaeHiddenEncoder(dec_extra.output, 512, 3136, name='cae_d_hid')
    resh = T.reshape(dec_l5.output, (b,64,7,7), ndim=4)
    dec_l6 = CaeConvolutionDecoder(resh, conv3_f, conv3_s, (b,64,7,7), name='cae_d1', border_mode='full')
    dec_l7 = CaeConvolutionDecoder(dec_l6.output, conv2_f, conv2_s, (b,64,18,18), name='cae_d2', border_mode='full')
    dec_l8 = CaeConvolutionDecoder(dec_l7.output, conv1_f, conv1_s, (b,32,84,84), name='cae_d3', border_mode='half') ## HACK: border_mode should be 'full' but the reslting dimensions are wrong
    o_layer = CaeOutputLayer(phi, dec_l8.output)
    layers = [enc_l1, enc_l2, enc_l3, enc_l4, enc_extra, dec_extra, dec_l5, dec_l6, dec_l7, dec_l8]

    theta = list( itertools.chain( *[layer.theta for layer in layers] ) )

    return o_layer, theta, dec_l7.output

def buildPredictionNet(phi, phi_n, batchSize, parameters): 
    layers = []
    chanels = 4
    b = batchSize

    filter_shape1 = (32, 4, 8, 8)
    filter_shape2 = (64, 32, 4, 4)
    filter_shape3 = (64, 64, 3, 3)

    ## Encoder
    enc_l1 = CaeConvolutionEncoder(phi, filter_shape1, (4,4), name='cae_e1', parameters=parameters[0])
    enc_l2 = CaeConvolutionEncoder(enc_l1.output, filter_shape2, (2,2), name='cae_e2', parameters=parameters[1])
    enc_l3 = CaeConvolutionEncoder(enc_l2.output, filter_shape3, (1,1), name='cae_e3', parameters=parameters[2])
    enc_l4 = CaeHiddenEncoder(T.flatten(enc_l3.output, outdim=2), 3136, 512, name='cae_e_hid', parameters=parameters[3]) 
    enc_out = enc_l4.output

    ## Decoder 
    dec_l5 = CaeHiddenEncoder(enc_out, 512, 3136, name='cae_d_hid')
    resh = T.reshape(dec_l5.output, (b,64,7,7), ndim=4)
    dec_l6 = CaeConvolutionDecoder(resh, filter_shape3, (1,1), (b,64,7,7), name='cae_d1', border_mode='full')
    dec_l7 = CaeConvolutionDecoder(dec_l6.output, filter_shape2, (2,2), (b,64,18,18), name='cae_d2', border_mode='full')
    dec_l8 = CaeConvolutionDecoder(dec_l7.output, filter_shape1, (4,4), (b,32,84,84), name='cae_d3', border_mode='half') ## HACK: border_mode should be 'full' but the reslting dimensions are wrong
    o_layer = CaeOutputLayer(phi_n, dec_l8.output)
    layers = [enc_l1, enc_l2, enc_l3, enc_l4, dec_l5, dec_l6, dec_l7, dec_l8]

    theta = list( itertools.chain( *[layer.theta for layer in layers] ) )

    return o_layer, theta

"""
def buildCaeLasagne(phi, batchSize): 
    l_in = InputLayer((None,4,84,84),input=phi)
    c1 = Conv2DLayer(l_in, num_filters=32, filter_size=8, stride=4, nonlinearity=rectify)
    c2 = Conv2DLayer(c1, num_filters=64, filter_size=4, stride=2, nonlinearity=rectify)
    c3 = Conv2DLayer(c2, num_filters=64, filter_size=3, stride=1, nonlinearity=rectify)
    d1 = DenseLayer(c3, num_units=512, nonlinearity=rectify)
    l_out = InverseLayer(d1, c1)
    help(l_out)
    o_layer = CaeOutputLayer(phi, l_out.get_output())

    return o_layer, l_out.get_all_params()
"""