from weightInitialization import zeros
import theano.tensor as T

def nature(p, g, lr, mom, rms_decay, rms_const): 
    shape = p.get_value(borrow=True).shape

    acc_mom = zeros(name='acc_mom', shape=shape)
    acc_mom_new = mom * acc_mom + (1-mom) * g

    acc_ms = zeros(name='acc_ms', shape=shape)
    acc_ms_new = rms_decay * acc_ms + (1-rms_decay) * g ** 2

    rms = T.sqrt(acc_ms_new-acc_mom**2+rms_const)
    p_new = p - lr * (g/rms)

    return [(acc_mom, acc_mom_new), (acc_ms, acc_ms_new), (p, p_new)]

def RMSprop(p, g, lr, mom, rms_decay, rms_const):
    shape = p.get_value(borrow=True).shape  
    """
    acc_ms = zeros(name='acc_ms', shape=shape)
    acc_ms_new = rms_decay * acc_ms + (1-rms_decay) * g ** 2 

    delta = - (lr * g)/(T.sqrt(acc_ms_new+rms_const))

    p_new = p + delta

    return [(acc_ms, acc_ms_new), (p, p_new)]
    """
    acc_ms = zeros(name='acc_ms', shape=shape)
    acc_ms_new = rms_decay * acc_ms + (1-rms_decay) * g ** 2 

    delta = - (lr * g)/(T.sqrt(acc_ms_new+rms_const))

    p_new = p + delta

    return [(acc_ms, acc_ms_new), (p, p_new)]

def momentum(p, g, lr, mom, rms_decay, rms_const): 
    shape = p.get_value(borrow=True).shape 

    velocity = zeros(name='acc_mom', shape=shape)
    velocity_new = mom * velocity - lr * g

    p_new = p + velocity_new

    return [(velocity, velocity_new), (p, p_new)]

def SGD(p, g, lr, mom, rms_decay, rms_const):
    p_new = p - lr * g
    return [(p, p_new)]

def RMSprop_momentum(p, g, lr, mom, rms_decay, rms_const): 
    shape = p.get_value(borrow=True).shape 

    acc_ms = zeros(name='acc_ms', shape=shape)
    acc_ms_new = rms_decay * acc_ms + (1-rms_decay) * g ** 2 

    delta = - lr * g/(T.sqrt(acc_ms_new+rms_const))

    velocity = zeros(name='acc_mom', shape=shape)
    velocity_new = mom * velocity + delta

    p_new = p + velocity_new

    return [(acc_ms, acc_ms_new), (p, p_new)]


def optimize(theta, grads, ot, lr=0.00025, mom=0.9, rms_decay=0.9, rms_const=0.00001): 
    if ot == 'nature': 
        uFn = nature
    elif ot == 'RMSprop': 
        uFn = RMSprop
    elif ot == 'momentum': 
        uFn = momentum
    elif ot == 'SGD': 
        uFn = SGD
    elif ot == 'RMSprop_momentum': 
        uFn = RMSprop_momentum
    else: 
        raise ValueError("Unrecognized optimization type: \"%s\"" % ot)

    updates = []
    for p, g in zip(theta, grads): 
        updates.extend(uFn(p, g, lr, mom, rms_decay, rms_const))

    return updates
