import numpy as np

import theano

## Replay memory should be moved to gpu, this can help:
##  http://stackoverflow.com/questions/15917849/how-can-i-assign-update-subset-of-tensor-shared-variable-in-theano
##          comment: set_subtensor have not worked in previoues attemts -> it fails quietly if dimentions are correct and the values are not copied 

class ReplayMemory:

    def __init__(self, batchSize, maxSize, ims = 84, imc = 4):
        print "Initiating replay memory"
        self.b = batchSize
        self.maxSize = maxSize
        self.size = 0 

        self.D_phi   = np.zeros((maxSize, imc, ims, ims), dtype='uint8')
        self.D_a     = np.zeros((maxSize,), dtype='int8')
        self.D_r     = np.zeros((maxSize,), dtype=theano.config.floatX)
        self.D_t     = np.zeros((maxSize,), dtype='int8')


    def store(self, phi, a, r, t): 
        # print "Input phi shape:",phi.shape
        # print "Replay memory shape:",self.D_phi.shape
        i = self.size%self.maxSize
        # print "Storing at index=", i
        np.copyto(self.D_phi[i,:], phi)
        self.D_a[i] = a
        self.D_r[i] = r
        self.D_t[i] = t

        self.size += 1

    def _indices(self): 
        #indices = np.random.choice(min(self.size+1, self.maxSize-1) - 1, self.b, replace=True) 

        _max = min(self.size, self.maxSize-1)
        indices = np.zeros((self.b))
        for i in range(self.b): 
            index = np.random.randint(_max)
            while self.D_t[index]: 
                index = np.random.randint(_max)
            indices[i] = index
        return indices.astype('int')


    def sample(self):
        ## The last added state does not have correct next state -> with full repmem next state will be a random state
        indices = self._indices()

        mode = 'wrap'
        b_phi = np.take(self.D_phi, indices, axis=0, mode=mode)
        b_a = np.take(self.D_a, indices, mode=mode)
        b_r = np.take(self.D_r, indices, mode=mode)
        b_phi_n = np.take(self.D_phi, (indices + 1), axis=0, mode=mode)
        b_t = np.take(self.D_t, indices, mode=mode)
        return b_phi, b_a, b_r, b_phi_n, b_t


    def save(self, fname): 
        kwds = {'phi':self.D_phi,
                'a': self.D_a, 
                'r': self.D_r, 
                't': self.D_t}
        np.savez(fname, **kwds )

    def load(self, fname): 
        npzfile = np.load(fname)

        if True: 
            print 'D_a: min=%i, max=%i' % (np.min, np.max())

        self.D_phi   = npzfile['phi']
        self.D_a     = npzfile['a']
        self.D_r     = npzfile['r']
        self.D_t     = npzfile['t']


    def reset(self): 
        self.size = 0

if __name__ == '__main__':
    """
    z = np.zeros((1,4,84,84), dtype='uint8')
    o = z + 1 
    rm = ReplayMemory(32, 100)

    for i in range(1, 51): 
        rm.store(o, i, 0, 0)

    for i in range(51,100): 
        rm.store(z, i, 0, 1)

    for i in range(10): 
        _, a, _, _, _ = rm.sample()
        print np.min(a), np.max(a)
    """
    from time import time

    z = np.zeros((1,4,84,84), dtype='uint8')
    rm = ReplayMemory(32, 250000)

    s1 = time()
    for i in range(250000): 
        rm.store(z, 0, 0, 0)
    e1 = time()

    print "Storing 250K samples took %f min" % ((e1-s1)/60)

    s2 = time()
    for i in range(1000): 
        temp = rm._indices()
    e2 = time()

    t2 = (e2-s2)/60
    print "1K calls methord:_indices %f min" % t2
    print "Total time for 5M frames is %f" % (t2*5000)

    s3 = time()
    for i in range(1000): 
        temp = rm.sample()
    e3= time()    

    t3 = (e3-s3)/60
    print "1K calls methord:sample %f min" % t3
    print "Total time for 5M frames is %f" % (t3*5000)