from os.path import join, abspath
import os
import pickle
import numpy as np
from scipy.signal import convolve2d


template = """
\\begin{tikzpicture}
\\begin{axis}[
    title={Something},
    xlabel={Steps},
    ylabel={Game Score},
    xmin=0, xmax=100,
    ymin=0, ymax=120,
    xtick={0, 125000, 250000, 375000, 500000, 625000, 750000, 875000, 1000000, 1125000, 1250000, 1375000, 1500000, 1625000, 1750000, 1875000, 2000000, 2125000, 2250000, 2375000, 2500000, 2625000, 2750000, 2875000, 3000000, 3125000, 3250000, 3375000, 3500000, 3625000, 3750000, 3875000, 4000000, 4125000, 4250000, 4375000, 4500000, 4625000, 4750000, 4875000, 5000000},
    ytick={0,20,40,60,80,100,120},
    legend pos=north west,
    ymajorgrids=true,
    grid style=dashed,  
]
 
%s
 
\\end{axis}
\\end{tikzpicture}
"""

plotTemplate = """
\\addplot[
    mark=square,
    ]
    coordinates {%s};
    \legend{%s}
"""

def movingAverage(v, n=5, mode='valid'): 
    print v.shape
    print v[::-1].shape
    print np.zeros((n-1, v.shape[1])).shape
    v = np.concatenate((v[::-1], np.zeros((n-1, v.shape[1]))), axis=0)
    ra = convolve2d(v, np.ones((n,1))/n, mode=mode)
    return ra[::-1]



def resPaths(folder): 
    resPaths = []
    for root, dirs, files in os.walk(folder, topdown=False):
        resPaths.extend([ abspath(join(root, name)) for name in files if name.endswith('res.dat') ])
    return resPaths

def load(path): 
    with open(path, 'rb') as f: 
        data = pickle.load(f)
    return data

def selectSteps(paths): 
    data = load(paths[0])
    steps = [ t[0][:-4] for t in data]
    return steps

def selectLabels(pahts): 
    return [ path.split('/')[-2] for path in paths]

def selectRes(paths): 
    res = []
    for path in paths: 
        data = load(path)
        res.append( np.array([ t[1] for t in data]) ) 
        print 'Data from %s, has length %i' % (path, len(res[-1]))
    return np.stack(res).T

def averageRes(labels, res): 
    default = ('', 0, [])
    dic = {}
    for i, l in enumerate(labels): 
        k = l[0:2]
        r = res[:,i]
        v = dic.get(k, ('', []))
        v[1].append(r)
        dic[k] = (l, v[1])
    items = dic.items()
    items.sort()
    labels = []
    res = []
    for _, v in items: 
        n = len(v[1])
        labels.append(str(n) + 'x_' + v[0])
        r = np.stack(v[1])
        r = np.mean(r, axis=0)
        res.append(r)
    res = np.stack(res).T
    return labels, res


def toCsv(fName, labels, steps, res): 
    labels = ['Steps'] + labels
    sep = ', '
    strings = [sep.join(labels)] 
    for s, r in zip(steps, res): 
        s = s[4:]
        strings.append(s + sep + sep.join([str(r_) for r_ in r]))
    string = '\n'.join(strings)
    with open(fName, 'w') as f: 
        f.writelines(string)

def toLatex(fName, labels, steps, resMat): 
    resMat = resMat.T
    plotStrings = []
    for l, resVec in zip(labels, resMat): 
        corString = ''.join( '(%s,%s)' % (s[4:], r) for s, r in zip(steps, resVec))
        plotString = plotTemplate % (corString, l)
        plotStrings.append(plotString)
    latex = template % ''.join(plotStrings)
    with open(fName, 'w') as f: 
        f.writelines(latex)


if __name__ == '__main__':
    paths = resPaths('.')
    paths.sort()
    labels = selectLabels(paths)
    steps = selectSteps(paths)
    res = selectRes(paths)
    labels, res = averageRes(labels, res)
    toCsv('results.csv', labels, steps, res)
    toCsv('results_avg5.csv', labels, steps, movingAverage(res, 5))
    #toLatex('latex.tex', labels, steps, res)
