__author__ = 'torstein'

import theano
import numpy as np

## Different initialization distributions
def zeros(name, shape): 
    W = np.zeros(shape, dtype=theano.config.floatX)
    return theano.shared(W, name=name, borrow=True)

def uniformDist(name, bound, shape, rng=np.random):
    W = np.asarray(
            rng.uniform(low=-bound, high=bound, size=shape),
            dtype=theano.config.floatX)
    return theano.shared(W, name=name, borrow=True)

def normalDist(name, shape, mean=0, std=1, rng=np.random): 
    W = np.asarray(
            rng.normal(mean, std, shape), 
            dtype=theano.config.floatX)
    return theano.shared(W, name=name, borrow=True)



## Initialization of convolutional weights
def initConv(name, shape):
    bound = 1.0 / (np.sqrt(shape[1]*shape[2]*shape[3]))
    return uniformDist(name, bound, shape)

def initConvB(name, shape): 
    bound = 1.0 / (np.sqrt(shape[1]*shape[2]*shape[3]))
    return uniformDist(name, bound, shape[0])



## Initialization of linear layer weights
def initW(name, n_in, n_out):
    shape = (n_in, n_out)
    bound = 1.0 / (np.sqrt(n_in))
    return uniformDist(name, bound, shape)

def initB(name, n_in, n_out):
    shape = (n_out,)
    bound = 1.0 / (np.sqrt(n_in))
    return uniformDist(name, bound, n_out)



## Misc
def copyTensor(t, name='copy_', prefix=True):
    temp = name + t.name if prefix else name
    return theano.shared( value = t.get_value(borrow=False), name=temp)

def chain(layout):
    l1 = layout[:-1]
    l2 = layout[1: ]
    return zip(l1,l2)